from __future__ import annotations

import logging
import os

import pandas as pd
import pytest

from napistu import consensus, indices, sbml_dfs_core, source
from napistu.constants import (
    BQB,
    IDENTIFIERS,
    SBML_DFS,
    SBML_DFS_METADATA,
    SBML_DFS_SCHEMA,
    SCHEMA_DEFS,
    SOURCE_SPEC,
)
from napistu.identifiers import Identifiers
from napistu.ingestion import sbml
from napistu.modify.cofactors import drop_cofactors

test_path = os.path.abspath(os.path.join(__file__, os.pardir))
test_data = os.path.join(test_path, "test_data")


def test_reduce_to_consensus_ids():
    sbml_path = os.path.join(test_data, "R-HSA-1237044.sbml")

    # test aggregating by IDs, by moving from compartmentalized_species -> species

    sbml_model = sbml.SBML(sbml_path)
    comp_species_df = sbml_model._define_cspecies()
    comp_species_df.index.names = [SBML_DFS.S_ID]
    consensus_species, species_lookup = consensus._reduce_to_consensus_ids(
        comp_species_df,
        {
            SCHEMA_DEFS.PK: SBML_DFS.S_ID,
            SCHEMA_DEFS.ID: SBML_DFS.S_IDENTIFIERS,
            SCHEMA_DEFS.TABLE: SBML_DFS.SPECIES,
        },
    )

    assert isinstance(consensus_species, pd.DataFrame)
    assert consensus_species.shape == (18, 4)
    assert isinstance(species_lookup, pd.Series)
    assert species_lookup.size == 23


def test_consensus():
    pw_index = indices.PWIndex(os.path.join(test_data, SOURCE_SPEC.PW_INDEX_FILE))
    sbml_dfs_dict = consensus.construct_sbml_dfs_dict(pw_index)

    consensus_model = consensus.construct_consensus_model(sbml_dfs_dict, pw_index)
    assert consensus_model.species.shape == (38, 3)
    assert consensus_model.reactions.shape == (30, 4)
    assert consensus_model.reaction_species.shape == (137, 4)

    consensus_model.validate()

    consensus_model = drop_cofactors(consensus_model)
    assert consensus_model.species.shape == (30, 3)
    assert consensus_model.reaction_species.shape == (52, 4)
    # update reaction_species.shape after more cofactors identified

    consensus_model.validate()


def test_source_tracking():
    # create input data
    table_schema = {SCHEMA_DEFS.SOURCE: "source_var", SCHEMA_DEFS.PK: "primary_key"}

    # define existing sources and the new_id entity they belong to
    # here, we are assuming that each model has a blank source object
    # as if it came from a non-consensus model
    agg_tbl = pd.DataFrame(
        {
            "new_id": [0, 0, 1, 1],
        }
    )
    agg_tbl[table_schema[SCHEMA_DEFS.SOURCE]] = source.Source.empty()

    # define new_ids and the models they came from
    # these models will be matched to the pw_index to flush out metadata
    lookup_table = pd.DataFrame(
        {
            "new_id": [0, 0, 1, 1],
            SOURCE_SPEC.MODEL: [
                "R-HSA-1237044",
                "R-HSA-425381",
                "R-HSA-1237044",
                "R-HSA-425381",
            ],
        }
    )

    # use an existing pw_index since pw_index currently checks for the existence of the source file
    pw_index = indices.PWIndex(os.path.join(test_data, "pw_index.tsv"))

    # test create source table
    source_table = source.create_source_table(lookup_table, table_schema, pw_index)
    assert source_table["source_var"][0].source.shape == (2, 8)

    # test create_consensus_sources
    consensus_sources = consensus._create_consensus_sources(
        agg_tbl, lookup_table, table_schema, pw_index
    )
    assert consensus_sources[0].source.shape == (2, 8)

    # lets add a model which does not have a reference in the pw_index
    invalid_lookup_table = pd.DataFrame(
        {
            "new_id": [0, 0, 1, 1],
            SOURCE_SPEC.MODEL: [
                "R-HSA-1237044",
                "R-HSA-425381",
                "R-HSA-1237044",
                "typo",
            ],
        }
    )

    # expect a ValueError when the model is not found
    with pytest.raises(ValueError) as _:
        source.create_source_table(invalid_lookup_table, table_schema, pw_index)

    # now we will aggregate the consensus model above with a new single model (which has some
    # overlapping entries with the consensusd (id 1) and some new ids (id 2)

    agg_tbl2 = pd.DataFrame(
        {
            "new_id": [0, 1, 1, 2],
        }
    )

    agg_tbl2[table_schema[SCHEMA_DEFS.SOURCE]] = consensus_sources.tolist() + [
        source.Source.empty() for i in range(0, 2)
    ]

    lookup_table2 = pd.DataFrame(
        {
            "new_id": [0, 1, 1, 2],
            # the model for the first two entries should really correspond to the "consensus"
            # but since this is not a file I will stub with one of the pw_index entries
            "model": [
                "R-HSA-1247673",
                "R-HSA-1247673",
                "R-HSA-1475029",
                "R-HSA-1475029",
            ],
        }
    )

    source_table = source.create_source_table(lookup_table2, table_schema, pw_index)
    assert source_table.shape == (3, 1)
    assert [
        source_table["source_var"][i].source.shape
        for i in range(0, source_table.shape[0])
    ] == [(1, 8), (2, 8), (1, 8)]

    consensus_sources = consensus._create_consensus_sources(
        agg_tbl2, lookup_table2, table_schema, pw_index
    )
    assert [
        consensus_sources[i].source.shape for i in range(0, consensus_sources.shape[0])
    ] == [(3, 8), (4, 8), (1, 8)]


def test_passing_entity_data():

    pw_index = indices.PWIndex(os.path.join(test_data, "pw_index.tsv"))
    sbml_dfs_dict = consensus.construct_sbml_dfs_dict(pw_index)

    for model in list(sbml_dfs_dict.keys())[0:3]:
        sbml_dfs_dict[model].add_species_data(
            "my_species_data",
            sbml_dfs_dict[model]
            .species.iloc[0:5]
            .assign(my_species_data_var="testing")["my_species_data_var"]
            .to_frame(),
        )
        sbml_dfs_dict[model].add_reactions_data(
            "my_reactions_data",
            sbml_dfs_dict[model]
            .reactions.iloc[0:5]
            .assign(my_reactions_data_var1="testing")
            .assign(my_reactions_data_var2="testing2")[
                ["my_reactions_data_var1", "my_reactions_data_var2"]
            ],
        )

    # create a consensus with perfect merges of overlapping id-table-variable values
    # i.e., when combined all merged entries have the same attributes
    consensus_model = consensus.construct_consensus_model(sbml_dfs_dict, pw_index)

    assert len(consensus_model.species_data) == 1
    assert consensus_model.species_data["my_species_data"].shape == (10, 1)
    assert len(consensus_model.reactions_data) == 1
    assert consensus_model.reactions_data["my_reactions_data"].shape == (14, 2)

    # add different tables from different models
    for model in list(sbml_dfs_dict.keys())[3:5]:
        sbml_dfs_dict[model].add_species_data(
            "my_other_species_data",
            sbml_dfs_dict[model]
            .species.iloc[0:5]
            .assign(my_species_data="testing")["my_species_data"]
            .to_frame(),
        )

    consensus_model = consensus.construct_consensus_model(sbml_dfs_dict, pw_index)
    assert len(consensus_model.species_data) == 2

    # create a case where reactions will be merged and the same reaction
    # in different models has a different value for its reactions_data
    minimal_pw_index = pw_index
    minimal_pw_index.index = minimal_pw_index.index.iloc[0:2]

    # Since we're working with a DataFrame, we can use loc to update the file value directly
    minimal_pw_index.index.loc[1, SOURCE_SPEC.FILE] = minimal_pw_index.index.loc[
        0, SOURCE_SPEC.FILE
    ]

    duplicated_sbml_dfs_dict = consensus.construct_sbml_dfs_dict(minimal_pw_index)
    # explicitely define the order we'll loop through models so that
    # the position of a model can be used to set mismatching attributes
    # for otherwise identical models
    model_order = list(duplicated_sbml_dfs_dict.keys())

    for model in duplicated_sbml_dfs_dict.keys():
        model_index = model_order.index(model)

        duplicated_sbml_dfs_dict[model].add_reactions_data(
            "my_mismatched_data",
            duplicated_sbml_dfs_dict[model]
            .reactions.iloc[0:5]
            .assign(my_reactions_data_var1=model)["my_reactions_data_var1"]
            .to_frame()
            .assign(numeric_var=[x + model_index for x in range(0, 5)])
            .assign(bool_var=[x + model_index % 2 == 0 for x in range(0, 5)]),
        )

    # assign reversibility is True for one model to
    # confirm that reversibility trumps irreversible
    # when merging reactions with identical stoichiometry but
    # different reversibility attributes

    duplicated_sbml_dfs_dict["R-HSA-1237044"].reactions = duplicated_sbml_dfs_dict[
        "R-HSA-1237044"
    ].reactions.assign(r_isreversible=True)

    consensus_model = consensus.construct_consensus_model(
        duplicated_sbml_dfs_dict, pw_index
    )
    assert consensus_model.reactions_data["my_mismatched_data"].shape == (5, 3)
    assert consensus_model.reactions[SBML_DFS.R_ISREVERSIBLE].eq(True).all()


def test_report_consensus_merges_reactions(tmp_path, model_source_stub):
    # Create two minimal SBML_dfs objects with a single reaction each, same r_id
    r_id = "R00000001"
    reactions = pd.DataFrame(
        {
            SBML_DFS.R_NAME: ["rxn1"],
            SBML_DFS.R_IDENTIFIERS: [None],
            SBML_DFS.R_SOURCE: [None],
            SBML_DFS.R_ISREVERSIBLE: [False],
        },
        index=[r_id],
    )
    reactions.index.name = SBML_DFS.R_ID
    reaction_species = pd.DataFrame(
        {
            SBML_DFS.R_ID: [r_id],
            SBML_DFS.SC_ID: ["SC0001"],
            SBML_DFS.STOICHIOMETRY: [1],
            SBML_DFS.SBO_TERM: ["SBO:0000459"],
        },
        index=["RSC0001"],
    )
    reaction_species.index.name = SBML_DFS.RSC_ID
    compartmentalized_species = pd.DataFrame(
        {
            SBML_DFS.SC_NAME: ["A [cytosol]"],
            SBML_DFS.S_ID: ["S0001"],
            SBML_DFS.C_ID: ["C0001"],
            SBML_DFS.SC_SOURCE: [None],
        },
        index=["SC0001"],
    )
    compartmentalized_species.index.name = SBML_DFS.SC_ID
    species = pd.DataFrame(
        {
            SBML_DFS.S_NAME: ["A"],
            SBML_DFS.S_IDENTIFIERS: [None],
            SBML_DFS.S_SOURCE: [None],
        },
        index=["S0001"],
    )
    species.index.name = SBML_DFS.S_ID
    compartments = pd.DataFrame(
        {
            SBML_DFS.C_NAME: ["cytosol"],
            SBML_DFS.C_IDENTIFIERS: [None],
            SBML_DFS.C_SOURCE: [None],
        },
        index=["C0001"],
    )
    compartments.index.name = SBML_DFS.C_ID
    sbml_dict = {
        SBML_DFS.COMPARTMENTS: compartments,
        SBML_DFS.SPECIES: species,
        SBML_DFS.COMPARTMENTALIZED_SPECIES: compartmentalized_species,
        SBML_DFS.REACTIONS: reactions,
        SBML_DFS.REACTION_SPECIES: reaction_species,
    }
    sbml1 = sbml_dfs_core.SBML_dfs(
        sbml_dict, model_source_stub, validate=False, resolve=False
    )
    sbml2 = sbml_dfs_core.SBML_dfs(
        sbml_dict, model_source_stub, validate=False, resolve=False
    )
    sbml_dfs_dict = {"mod1": sbml1, "mod2": sbml2}

    # Create a lookup_table that merges both reactions into a new_id
    lookup_table = pd.DataFrame(
        {
            SOURCE_SPEC.MODEL: ["mod1", "mod2"],
            SBML_DFS.R_ID: [r_id, r_id],
            "new_id": ["merged_rid", "merged_rid"],
        }
    )
    # Use the reactions schema
    table_schema = SBML_DFS_SCHEMA.SCHEMA[SBML_DFS.REACTIONS]

    # Call the function and check that it runs and the merge_labels are as expected
    consensus._report_consensus_merges(
        lookup_table.set_index([SOURCE_SPEC.MODEL, SBML_DFS.R_ID])[
            "new_id"
        ],  # this is a Series with name 'new_id'
        table_schema,
        sbml_dfs_dict=sbml_dfs_dict,
        n_example_merges=1,
    )
    # No assertion: this is a smoke test to ensure the Series output is handled without error


def test_build_consensus_identifiers_handles_merges_and_missing_ids():

    # Three entities:
    # - 'A' with identifier X
    # - 'B' with no identifiers
    # - 'C' with identifier X (should merge with 'A')
    df = pd.DataFrame(
        {
            SBML_DFS.S_ID: ["A", "B", "C"],
            SBML_DFS.S_IDENTIFIERS: [
                Identifiers(
                    [
                        {
                            IDENTIFIERS.ONTOLOGY: "test",
                            IDENTIFIERS.IDENTIFIER: "X",
                            IDENTIFIERS.BQB: BQB.IS,
                        }
                    ]
                ),
                Identifiers([]),
                Identifiers(
                    [
                        {
                            IDENTIFIERS.ONTOLOGY: "test",
                            IDENTIFIERS.IDENTIFIER: "X",
                            IDENTIFIERS.BQB: BQB.IS,
                        }
                    ]
                ),
            ],
        }
    ).set_index(SBML_DFS.S_ID)

    schema = SBML_DFS_SCHEMA.SCHEMA[SBML_DFS.SPECIES]

    indexed_cluster, cluster_consensus_identifiers = (
        consensus._build_consensus_identifiers(df, schema)
    )

    # All entities should be assigned to a cluster
    assert set(indexed_cluster.index) == set(df.index)
    assert not indexed_cluster.isnull().any()
    # There should be a consensus identifier for each cluster
    assert set(cluster_consensus_identifiers.index) == set(indexed_cluster.values)

    # Entities 'A' and 'C' should be merged (same cluster)
    assert indexed_cluster.loc["A"] == indexed_cluster.loc["C"]
    # Entity 'B' should be in a different cluster
    assert indexed_cluster.loc["B"] != indexed_cluster.loc["A"]

    # The consensus identifier for the merged cluster should include identifier X
    merged_cluster_id = indexed_cluster.loc["A"]
    ids_obj = cluster_consensus_identifiers.loc[
        merged_cluster_id, schema[SCHEMA_DEFS.ID]
    ]
    assert "X" in ids_obj.df[IDENTIFIERS.IDENTIFIER].values

    # The consensus identifier for the entity with no identifiers should be empty
    noid_cluster_id = indexed_cluster.loc["B"]
    ids_obj_noid = cluster_consensus_identifiers.loc[
        noid_cluster_id, schema[SCHEMA_DEFS.ID]
    ]
    assert hasattr(ids_obj_noid, "df")
    assert ids_obj_noid.df.shape[0] == 0


def test_consensus_round_trip_consistency(
    pw_index_metabolism, sbml_dfs_dict_metabolism, sbml_dfs_metabolism
):
    """
    Test that round-trip conversion between different data structures produces consistent results.

    This test verifies that:
    1. sbml_dfs_dict -> sbml_dfs_list -> (sbml_dfs_dict, pw_index) -> consensus_model
    2. pw_index -> sbml_dfs_dict -> consensus_model

    Both paths should produce the same consensus model.
    """

    # Path 1: pw_index -> sbml_dfs_dict -> consensus_model was already created in conftest.py
    pw_index = pw_index_metabolism
    sbml_dfs_dict = sbml_dfs_dict_metabolism
    sbml_dfs = sbml_dfs_metabolism

    # Path 2: sbml_dfs_dict -> sbml_dfs_list -> (sbml_dfs_dict, pw_index) -> consensus_model
    sbml_dfs_list = list(sbml_dfs_dict.values())
    sbml_dfs_dict_from_list, pw_index_from_list = consensus.prepare_consensus_model(
        sbml_dfs_list
    )
    sbml_dfs_from_list = consensus.construct_consensus_model(
        sbml_dfs_dict_from_list, pw_index_from_list
    )

    # Verify both paths produce identical consensus models
    assert (
        sbml_dfs.species.shape == sbml_dfs_from_list.species.shape
    ), "Species tables should have same shape"
    assert (
        sbml_dfs.reactions.shape == sbml_dfs_from_list.reactions.shape
    ), "Reactions tables should have same shape"
    assert (
        sbml_dfs.reaction_species.shape == sbml_dfs_from_list.reaction_species.shape
    ), "Reaction species tables should have same shape"

    # Verify the pw_index objects are equivalent
    assert (
        pw_index.index.shape == pw_index_from_list.index.shape
    ), "PWIndex objects should have same shape"

    # Compare columns using set comparison
    assert set(pw_index.index.columns) == set(
        pw_index_from_list.index.columns
    ), "PWIndex objects should have same columns"

    # Compare the actual pw_index data, ignoring row/column order and dtype differences
    pd.testing.assert_frame_equal(
        pw_index.index,
        pw_index_from_list.index,
        check_like=True,  # Ignore the order of index & columns
        check_dtype=False,  # Ignore dtype differences (e.g., float64 vs object for date column)
    )

    # Verify the sbml_dfs_dict objects are equivalent
    assert len(sbml_dfs_dict) == len(
        sbml_dfs_dict_from_list
    ), "SBML_dfs dictionaries should have same number of entries"
    assert set(sbml_dfs_dict.keys()) == set(
        sbml_dfs_dict_from_list.keys()
    ), "SBML_dfs dictionaries should have same keys"

    # Compare the metadata of each paired SBML_dfs object
    for key in sbml_dfs_dict.keys():
        original_sbml_dfs = sbml_dfs_dict[key]
        reconstructed_sbml_dfs = sbml_dfs_dict_from_list[key]

        # Compare metadata using pandas testing
        pd.testing.assert_frame_equal(
            original_sbml_dfs.metadata[SBML_DFS_METADATA.SBML_DFS_SOURCE].source,
            reconstructed_sbml_dfs.metadata[SBML_DFS_METADATA.SBML_DFS_SOURCE].source,
        )


def test_pre_consensus_compartment_check_compatible(
    sbml_dfs_dict_metabolism, pw_index_metabolism
):
    """Test that compatible models with overlapping compartments pass the check."""
    # This should not raise any errors or log warnings
    consensus._pre_consensus_compartment_check(
        sbml_dfs_dict_metabolism, pw_index_metabolism
    )


def test_pre_consensus_compartment_check_incompatible(
    caplog, minimal_valid_sbml_dfs, sbml_dfs
):
    """Test that incompatible models with non-overlapping compartments are detected."""

    # Create a pathway index with both models
    pw_index_data = pd.DataFrame(
        {
            SOURCE_SPEC.FILE: ["fake_model", "real_model"],
            SOURCE_SPEC.DATA_SOURCE: ["Test", "Reactome"],
            SOURCE_SPEC.ORGANISMAL_SPECIES: ["Homo sapiens", "Homo sapiens"],
            SOURCE_SPEC.PATHWAY_ID: ["fake_model", "real_model"],
            SOURCE_SPEC.NAME: ["Fake Pathway", "Real Pathway"],
            SOURCE_SPEC.DATE: ["2023-01-01", "2023-01-01"],
        }
    )

    # Create SBML_dfs dictionary with incompatible models
    # minimal_valid_sbml_dfs has no identifiers, so no merging will occur
    sbml_dfs_dict = {"fake_model": minimal_valid_sbml_dfs, "real_model": sbml_dfs}

    # Create pathway index object
    pw_index = indices.PWIndex(pw_index_data, validate_paths=False)

    # Capture log messages
    with caplog.at_level(logging.ERROR):
        consensus._pre_consensus_compartment_check(sbml_dfs_dict, pw_index)

    # Check that an error was logged about incompatible compartments
    assert len(caplog.records) > 0
    assert any("incompatible" in record.message.lower() for record in caplog.records)


def test_pre_consensus_ontology_check_compatible(sbml_dfs_dict_metabolism):
    """Test that compatible models with overlapping ontologies pass the check."""
    # This should not raise any errors or log warnings
    consensus._pre_consensus_ontology_check(sbml_dfs_dict_metabolism, SBML_DFS.SPECIES)


def test_pre_consensus_ontology_check_incompatible(
    caplog, minimal_valid_sbml_dfs, sbml_dfs
):
    """Test that incompatible models with non-overlapping ontologies are detected."""

    # Create a modified minimal SBML_dfs with fake ontology identifiers
    fake_sbml_dfs = minimal_valid_sbml_dfs.copy()

    # Create fake identifiers with a fake ontology
    fake_identifiers = Identifiers(
        [
            {
                IDENTIFIERS.ONTOLOGY: "FAKE_ONTOLOGY",
                IDENTIFIERS.IDENTIFIER: "FAKE_ID",
                IDENTIFIERS.BQB: BQB.IS,
            }
        ]
    )

    # Update the species identifiers to use the fake ontology
    fake_species = fake_sbml_dfs.species.copy()
    fake_species[SBML_DFS.S_IDENTIFIERS] = [fake_identifiers]
    fake_sbml_dfs.species = fake_species

    # Create SBML_dfs dictionary with incompatible models
    sbml_dfs_dict = {"fake_model": fake_sbml_dfs, "real_model": sbml_dfs}

    # Capture log messages
    with caplog.at_level(logging.ERROR):
        consensus._pre_consensus_ontology_check(sbml_dfs_dict, SBML_DFS.SPECIES)

    # Check that an error was logged about incompatible ontologies
    assert len(caplog.records) > 0
    assert any("ontologies" in record.message.lower() for record in caplog.records)
    assert any(
        "incompatible" in record.message.lower() or "overlap" in record.message.lower()
        for record in caplog.records
    )


def test_consensus_species_merged(sbml_dfs_metabolism):
    """Test that species identifiers are properly merged during consensus building."""
    # Check for species merging; are defining IDs shared across species?
    species_ids = sbml_dfs_metabolism.get_identifiers(
        SBML_DFS.SPECIES, filter_by_bqb="defining"
    )
    assert all(
        species_ids.value_counts([IDENTIFIERS.ONTOLOGY, IDENTIFIERS.IDENTIFIER]) == 1
    )


def test_update_foreign_keys():
    """Test _update_foreign_keys function with various scenarios."""
    # Create test data
    agg_tbl = pd.DataFrame(
        {
            SOURCE_SPEC.MODEL: ["model_a", "model_a", "model_b", "model_b"],
            "entity_id": ["e1", "e2", "e3", "e4"],
            SBML_DFS.S_ID: ["s1", "s2", "s1", "s3"],
            "data": [10, 20, 30, 40],
        }
    ).set_index([SOURCE_SPEC.MODEL, "entity_id"])

    # Create FK lookup table (multiindex: model, old_id -> new_id)
    fk_lookup = pd.DataFrame(
        {"new_id": ["new_s1", "new_s2", "new_s1", "new_s3"]},
        index=pd.MultiIndex.from_tuples(
            [
                ("model_a", "s1"),
                ("model_a", "s2"),
                ("model_b", "s1"),
                ("model_b", "s3"),
            ],
            names=[SOURCE_SPEC.MODEL, SBML_DFS.S_ID],
        ),
    )

    # Create table schema
    table_schema = {SCHEMA_DEFS.FK: [SBML_DFS.S_ID], SCHEMA_DEFS.PK: "entity_id"}

    fk_lookup_tables = {SBML_DFS.S_ID: fk_lookup}

    # Test successful FK update
    result = consensus._update_foreign_keys(agg_tbl, table_schema, fk_lookup_tables)

    # Verify the foreign keys were updated correctly
    expected_species_ids = ["new_s1", "new_s2", "new_s1", "new_s3"]
    assert result[SBML_DFS.S_ID].tolist() == expected_species_ids
    assert result["data"].tolist() == [10, 20, 30, 40]  # Other data preserved


def test_update_foreign_keys_missing_key():
    """Test _update_foreign_keys raises ValueError for missing keys."""
    # Create test data with a key that won't be in the lookup table
    agg_tbl = pd.DataFrame(
        {
            SOURCE_SPEC.MODEL: ["model_a", "model_a"],
            "entity_id": ["e1", "e2"],
            SBML_DFS.S_ID: ["s1", "s_missing"],  # s_missing not in lookup
            "data": [10, 20],
        }
    ).set_index([SOURCE_SPEC.MODEL, "entity_id"])

    # Create FK lookup table (missing s_missing)
    fk_lookup = pd.DataFrame(
        {"new_id": ["new_s1"]},
        index=pd.MultiIndex.from_tuples(
            [("model_a", "s1")], names=[SOURCE_SPEC.MODEL, SBML_DFS.S_ID]
        ),
    )

    table_schema = {SCHEMA_DEFS.FK: [SBML_DFS.S_ID], SCHEMA_DEFS.PK: "entity_id"}

    fk_lookup_tables = {SBML_DFS.S_ID: fk_lookup}

    # Should raise ValueError for missing key
    with pytest.raises(
        ValueError,
        match=f"keys from agg_tbl are missing from the {SBML_DFS.S_ID} lookup table",
    ):
        consensus._update_foreign_keys(agg_tbl, table_schema, fk_lookup_tables)


def test_update_foreign_keys_multiple_fks():
    """Test _update_foreign_keys with multiple foreign keys using compartmentalized species."""
    # Create test data for compartmentalized species (has both s_id and c_id FKs)
    agg_tbl = pd.DataFrame(
        {
            SOURCE_SPEC.MODEL: ["model_a", "model_a"],
            SBML_DFS.SC_ID: ["sc1", "sc2"],
            SBML_DFS.S_ID: ["s1", "s2"],
            SBML_DFS.C_ID: ["c1", "c2"],
            "data": [10, 20],
        }
    ).set_index([SOURCE_SPEC.MODEL, SBML_DFS.SC_ID])

    # Create lookup tables for both FKs
    species_lookup = pd.DataFrame(
        {"new_id": ["new_s1", "new_s2"]},
        index=pd.MultiIndex.from_tuples(
            [("model_a", "s1"), ("model_a", "s2")],
            names=[SOURCE_SPEC.MODEL, SBML_DFS.S_ID],
        ),
    )

    compartment_lookup = pd.DataFrame(
        {"new_id": ["new_c1", "new_c2"]},
        index=pd.MultiIndex.from_tuples(
            [("model_a", "c1"), ("model_a", "c2")],
            names=[SOURCE_SPEC.MODEL, SBML_DFS.C_ID],
        ),
    )

    table_schema = {
        SCHEMA_DEFS.FK: [SBML_DFS.S_ID, SBML_DFS.C_ID],
        SCHEMA_DEFS.PK: SBML_DFS.SC_ID,
    }

    fk_lookup_tables = {
        SBML_DFS.S_ID: species_lookup,
        SBML_DFS.C_ID: compartment_lookup,
    }

    # Test updating both FKs
    result = consensus._update_foreign_keys(agg_tbl, table_schema, fk_lookup_tables)

    # Verify both FKs were updated
    assert result[SBML_DFS.S_ID].tolist() == ["new_s1", "new_s2"]
    assert result[SBML_DFS.C_ID].tolist() == ["new_c1", "new_c2"]
    assert result["data"].tolist() == [10, 20]  # Other data preserved


def test_construct_consensus_model_no_rxn_pathway_ids(
    sbml_dfs_dict_metabolism, pw_index_metabolism, sbml_dfs_metabolism
):
    """Test construct_consensus_model with no_rxn_pathway_ids parameter."""
    # Test with one pathway marked as no_rxn_pathway_ids
    no_rxn_pathway_ids = ["tca"]  # Select TCA cycle as no-reaction pathway

    # Create consensus model with no_rxn_pathway_ids
    consensus_model = consensus.construct_consensus_model(
        sbml_dfs_dict_metabolism,
        pw_index_metabolism,
        no_rxn_pathway_ids=no_rxn_pathway_ids,
    )

    assert consensus_model.reactions.shape[0] < sbml_dfs_metabolism.reactions.shape[0]
    # cleanup generally removes some unused cspecies, species, and in some cases compartments
    assert consensus_model.species.shape[0] < sbml_dfs_metabolism.species.shape[0]
    assert (
        consensus_model.compartmentalized_species.shape[0]
        < sbml_dfs_metabolism.compartmentalized_species.shape[0]
    )

    # Test with invalid pathway ID
    invalid_no_rxn_pathway_ids = ["invalid_pathway_id"]

    # Should raise ValueError when invalid pathway ID is provided
    with pytest.raises(ValueError) as exc_info:
        consensus.construct_consensus_model(
            sbml_dfs_dict_metabolism,
            pw_index_metabolism,
            no_rxn_pathway_ids=invalid_no_rxn_pathway_ids,
        )

    # Verify the error message mentions the invalid pathway ID
    assert "invalid_pathway_id" in str(exc_info.value)
    assert "not found in the pw_index" in str(exc_info.value)



"""
Utilities for runtime Python package inspection.

Provides functions to introspect installed Python packages and extract
source code, signatures, and metadata from functions, classes, and methods.
"""

import importlib
import inspect
from typing import Any, Dict, List, Optional, Tuple

from pydantic import BaseModel, Field


class FunctionInfo(BaseModel):
    """Information about a function."""

    source: Optional[str] = Field(None, description="Full source code")
    signature: Optional[str] = Field(
        None, description="Function signature with type hints"
    )
    docstring: Optional[str] = Field(None, description="Function docstring")
    file_path: Optional[str] = Field(
        None, description="Path to file containing the function"
    )
    line_number: Optional[int] = Field(None, description="Starting line number")

    @classmethod
    def from_function(cls, func: Any) -> "FunctionInfo":
        """
        Create FunctionInfo from a function object.

        Parameters
        ----------
        func : Any
            Function object to inspect

        Returns
        -------
        FunctionInfo
            Pydantic model containing function details
        """
        # Get source code
        try:
            source = inspect.getsource(func)
        except (OSError, TypeError):
            source = None

        # Get signature
        try:
            signature = str(inspect.signature(func))
        except (ValueError, TypeError):
            signature = None

        # Get docstring
        docstring = inspect.getdoc(func)

        # Get file location
        try:
            file_path = inspect.getfile(func)
            line_number = inspect.getsourcelines(func)[1]
        except (OSError, TypeError):
            file_path = None
            line_number = None

        return cls(
            source=source,
            signature=signature,
            docstring=docstring,
            file_path=file_path,
            line_number=line_number,
        )


class MethodInfo(BaseModel):
    """Information about a single method."""

    signature: Optional[str] = Field(None, description="Method signature")
    docstring: Optional[str] = Field(None, description="Method docstring")


class ClassInfo(BaseModel):
    """Information about a class."""

    name: str = Field(description="Class name")
    module: str = Field(description="Module where class is defined")
    docstring: Optional[str] = Field(None, description="Class docstring")
    file_path: Optional[str] = Field(
        None, description="Path to file containing the class"
    )
    line_number: Optional[int] = Field(None, description="Starting line number")
    init_signature: Optional[str] = Field(None, description="__init__ method signature")
    init_source: Optional[str] = Field(None, description="__init__ source code")
    methods: Dict[str, MethodInfo] = Field(
        default_factory=dict, description="Dictionary of methods"
    )

    @classmethod
    def from_class(cls, class_obj: Any, include_init: bool = True) -> "ClassInfo":
        """
        Create ClassInfo from a class object.

        Parameters
        ----------
        class_obj : Any
            Class object to inspect
        include_init : bool, optional
            Whether to include __init__ source code (default: True)

        Returns
        -------
        ClassInfo
            Pydantic model containing class details
        """
        # Get docstring
        docstring = inspect.getdoc(class_obj)

        # Get file location
        try:
            file_path = inspect.getfile(class_obj)
            line_number = inspect.getsourcelines(class_obj)[1]
        except (OSError, TypeError):
            file_path = None
            line_number = None

        # Get __init__ info
        init_signature = None
        init_source = None
        if hasattr(class_obj, "__init__"):
            init_method = class_obj.__init__
            try:
                init_signature = str(inspect.signature(init_method))
            except (ValueError, TypeError):
                pass

            if include_init:
                try:
                    init_source = inspect.getsource(init_method)
                except (OSError, TypeError):
                    pass

        # Get methods (excluding private unless __init__)
        methods = {}
        for name, method in inspect.getmembers(class_obj, inspect.isfunction):
            if not name.startswith("_") or name == "__init__":
                try:
                    sig = str(inspect.signature(method))
                except (ValueError, TypeError):
                    sig = None

                methods[name] = MethodInfo(
                    signature=sig, docstring=inspect.getdoc(method)
                )

        return cls(
            name=class_obj.__name__,
            module=class_obj.__module__,
            docstring=docstring,
            file_path=file_path,
            line_number=line_number,
            init_signature=init_signature,
            init_source=init_source,
            methods=methods,
        )


class MethodSourceInfo(BaseModel):
    """Detailed source information for a specific method."""

    method_name: str = Field(description="Name of the method")
    class_name: str = Field(description="Name of the containing class")
    source: Optional[str] = Field(None, description="Method source code")
    signature: Optional[str] = Field(None, description="Method signature")
    docstring: Optional[str] = Field(None, description="Method docstring")
    line_number: Optional[int] = Field(None, description="Starting line number")
    error: Optional[str] = Field(None, description="Error message if method not found")

    @classmethod
    def from_method(cls, class_obj: Any, method_name: str) -> "MethodSourceInfo":
        """
        Create MethodSourceInfo from a class and method name.

        Parameters
        ----------
        class_obj : Any
            Class object containing the method
        method_name : str
            Name of the method to extract

        Returns
        -------
        MethodSourceInfo
            Pydantic model containing method details
        """
        # Check if method exists
        if not hasattr(class_obj, method_name):
            return cls(
                method_name=method_name,
                class_name=class_obj.__name__,
                error=f"Method '{method_name}' not found on class '{class_obj.__name__}'",
            )

        method = getattr(class_obj, method_name)

        # Get source
        try:
            source = inspect.getsource(method)
        except (OSError, TypeError):
            source = None

        # Get signature
        try:
            signature = str(inspect.signature(method))
        except (ValueError, TypeError):
            signature = None

        # Get docstring
        docstring = inspect.getdoc(method)

        # Get line number
        try:
            line_number = inspect.getsourcelines(method)[1]
        except (OSError, TypeError):
            line_number = None

        return cls(
            method_name=method_name,
            class_name=class_obj.__name__,
            source=source,
            signature=signature,
            docstring=docstring,
            line_number=line_number,
        )


def import_object(
    full_path: str, package_name: str = "napistu"
) -> Tuple[Any, Optional[str]]:
    """
    Import and return an object from a dotted path.

    Parameters
    ----------
    full_path : str
        Dotted path to the object. Can be:
        - Short form: "network.create_consensus" (uses package_name)
        - Full form: "napistu.network.create_consensus" (ignores package_name)
        - Class form: "sbml_dfs_core.SBML_dfs" (uses package_name)
    package_name : str, optional
        Default package name if full_path doesn't include it (default: "napistu")

    Returns
    -------
    Tuple[Any, Optional[str]]
        (object, error_message) where error_message is None on success

    Examples
    --------
    >>> obj, error = import_object("network.create_consensus", "napistu")
    >>> if error:
    ...     print(f"Import failed: {error}")
    ... else:
    ...     print(f"Got object: {obj}")

    >>> cls, error = import_object("sbml_dfs_core.SBML_dfs", "napistu")
    >>> if not error:
    ...     print(f"Got class: {cls.__name__}")
    """
    # If the path doesn't start with a known package, prepend package_name
    if not full_path.startswith(package_name):
        full_path = f"{package_name}.{full_path}"

    # Split into module path and object name
    parts = full_path.split(".")

    # Try progressively shorter module paths, starting from longest
    # This handles cases like:
    # - napistu.network.create_consensus (module: napistu.network, obj: create_consensus)
    # - napistu.sbml_dfs_core.SBML_dfs (module: napistu.sbml_dfs_core, obj: SBML_dfs)
    last_error = None

    for i in range(len(parts), 0, -1):
        module_path = ".".join(parts[:i])
        obj_path = parts[i:]

        try:
            # Try to import the module
            module = importlib.import_module(module_path)

            # If no remaining path, return the module itself
            if not obj_path:
                return module, None

            # Navigate to the object through the remaining path
            obj = module
            for attr_name in obj_path:
                obj = getattr(obj, attr_name)

            return obj, None

        except ImportError as e:
            # Module doesn't exist, try shorter path
            last_error = e
            continue
        except AttributeError as e:
            # Module exists but doesn't have the attribute, try shorter path
            last_error = e
            continue

    # If we got here, nothing worked
    error_msg = f"Could not import '{full_path}'. "
    if last_error:
        error_msg += f"Last error: {str(last_error)}"
    else:
        error_msg += "Module or object not found."

    return None, error_msg


def list_installed_packages() -> List[Dict[str, str]]:
    """
    List all installed Python packages with their versions.

    Returns
    -------
    List[Dict[str, str]]
        List of package dictionaries, each containing:
        - name : str
            Package name
        - version : str
            Installed version
        Packages are sorted alphabetically by name.

    Examples
    --------
    >>> packages = list_installed_packages()
    >>> pandas_pkg = next((p for p in packages if p["name"] == "pandas"), None)
    >>> if pandas_pkg:
    ...     print(f"pandas version: {pandas_pkg['version']}")
    """
    try:
        # Python 3.8+ - preferred method
        from importlib.metadata import distributions

        packages = []
        for dist in distributions():
            try:
                name = dist.metadata["Name"]
                version = dist.metadata["Version"]
                if name:  # Skip packages without names
                    packages.append({"name": name, "version": version})
            except (KeyError, AttributeError):
                # Skip packages with missing metadata
                continue

        # Sort by name for consistent output
        packages.sort(key=lambda x: x["name"].lower())
        return packages

    except ImportError:
        # Fallback to pkg_resources for older Python versions
        import pkg_resources

        packages = []
        for dist in pkg_resources.working_set:
            packages.append({"name": dist.project_name, "version": dist.version})

        packages.sort(key=lambda x: x["name"].lower())
        return packages



from __future__ import annotations

import pandas as pd
import pytest

from napistu.network import data_handling, net_create


# Fixtures
@pytest.fixture
def mock_sbml_dfs():
    """Create a mock SBML_dfs object with test data."""

    class MockSBMLDfs:
        def __init__(self):
            self.species_data = {
                "test_table": pd.DataFrame(
                    {
                        "col1": [1, 2, 3],
                        "col2": ["a", "b", "c"],
                        "test_prefix_col": [4, 5, 6],
                    }
                ),
                "another_table": pd.DataFrame(
                    {"col3": [7, 8, 9], "col4": ["d", "e", "f"]}
                ),
            }
            self.reactions_data = {
                "reaction_table": pd.DataFrame(
                    {"rxn_col1": [10, 11, 12], "rxn_col2": ["g", "h", "i"]}
                )
            }

    return MockSBMLDfs()


@pytest.fixture
def test_entity_data():
    """Create a test data table."""
    return pd.DataFrame(
        {
            "attr1": [1, 2, 3],
            "attr2": ["a", "b", "c"],
            "test_prefix_attr": [4, 5, 6],
            "another_attr": [7, 8, 9],
        }
    )


def test_select_sbml_dfs_data_table(mock_sbml_dfs):
    """Test selecting data tables from SBML_dfs object."""
    # Test selecting specific species table
    result = data_handling._select_sbml_dfs_data_table(
        mock_sbml_dfs, "test_table", "species"
    )
    assert isinstance(result, pd.DataFrame)
    assert result.equals(mock_sbml_dfs.species_data["test_table"])

    # Test selecting reactions table
    result = data_handling._select_sbml_dfs_data_table(
        mock_sbml_dfs, "reaction_table", "reactions"
    )
    assert isinstance(result, pd.DataFrame)
    assert result.equals(mock_sbml_dfs.reactions_data["reaction_table"])

    # Test error cases
    with pytest.raises(ValueError, match="Invalid table_type"):
        data_handling._select_sbml_dfs_data_table(
            mock_sbml_dfs, table_type="invalid_type"
        )

    with pytest.raises(ValueError, match="Invalid table_name"):
        data_handling._select_sbml_dfs_data_table(
            mock_sbml_dfs, "invalid_table", "species"
        )

    # Test no data case
    mock_sbml_dfs.species_data = {}
    with pytest.raises(ValueError, match="No species data found"):
        data_handling._select_sbml_dfs_data_table(mock_sbml_dfs)

    # Test multiple tables without specifying name
    mock_sbml_dfs.species_data = {
        "table1": pd.DataFrame({"col1": [1]}),
        "table2": pd.DataFrame({"col2": [2]}),
    }
    with pytest.raises(
        ValueError, match="Expected a single species data table but found 2"
    ):
        data_handling._select_sbml_dfs_data_table(mock_sbml_dfs)


def test_select_data_table_attrs_basic(test_entity_data):
    """Test basic attribute selection from data table."""
    # Test single attribute as list
    result = data_handling._create_data_table_column_mapping(
        test_entity_data, ["attr1"]
    )
    assert isinstance(result, dict)
    assert result == {"attr1": "attr1"}

    # Test multiple attributes
    result = data_handling._create_data_table_column_mapping(
        test_entity_data, ["attr1", "attr2"]
    )
    assert isinstance(result, dict)
    assert result == {"attr1": "attr1", "attr2": "attr2"}

    # Test invalid attribute
    with pytest.raises(ValueError, match="following attributes were missing"):
        data_handling._create_data_table_column_mapping(
            test_entity_data, ["invalid_attr"]
        )


def test_select_data_table_attrs_advanced(test_entity_data):
    """Test advanced attribute selection features."""
    # Test dictionary renaming
    result = data_handling._create_data_table_column_mapping(
        test_entity_data, {"attr1": "new_name1", "attr2": "new_name2"}
    )
    assert isinstance(result, dict)
    assert result == {"attr1": "new_name1", "attr2": "new_name2"}

    # Test empty dictionary
    with pytest.raises(ValueError, match="No attributes found in the dictionary"):
        data_handling._create_data_table_column_mapping(test_entity_data, {})

    # Test invalid source columns
    with pytest.raises(ValueError, match="following source columns were missing"):
        data_handling._create_data_table_column_mapping(
            test_entity_data, {"invalid_attr": "new_name"}
        )

    # Test conflicting new column names
    with pytest.raises(
        ValueError, match="following new column names conflict with existing columns"
    ):
        data_handling._create_data_table_column_mapping(
            test_entity_data,
            {"attr1": "attr2"},  # trying to rename attr1 to attr2, which already exists
        )

    # Test None returns identity mapping for all columns
    result = data_handling._create_data_table_column_mapping(test_entity_data, None)
    assert isinstance(result, dict)
    expected = {col: col for col in test_entity_data.columns}
    assert result == expected

    # Test string pattern matching
    result = data_handling._create_data_table_column_mapping(
        test_entity_data, "test_prefix_.*"
    )
    assert isinstance(result, dict)
    assert result == {"test_prefix_attr": "test_prefix_attr"}


def test_create_graph_attrs_config():
    """Test creating graph attributes configuration with a single table and transformation."""
    # Test basic case with single table and transformation
    result = data_handling._create_graph_attrs_config(
        column_mapping={"col1": "col1"},
        data_type="species",
        table_name="test_table",
        transformation="identity",
    )

    expected = {
        "species": {
            "col1": {
                "table": "test_table",
                "variable": "col1",
                "trans": "identity",
            }
        }
    }
    assert result == expected

    # Test with column renaming
    result = data_handling._create_graph_attrs_config(
        column_mapping={"original_col": "new_col"},
        data_type="species",
        table_name="test_table",
        transformation="squared",
    )

    expected = {
        "species": {
            "new_col": {
                "table": "test_table",
                "variable": "original_col",
                "trans": "squared",
            }
        }
    }
    assert result == expected

    # Test with multiple columns but same table and transformation
    result = data_handling._create_graph_attrs_config(
        column_mapping={"col1": "col1", "col2": "renamed_col2"},
        data_type="species",
        table_name="test_table",
        transformation="identity",
    )

    expected = {
        "species": {
            "col1": {
                "table": "test_table",
                "variable": "col1",
                "trans": "identity",
            },
            "renamed_col2": {
                "table": "test_table",
                "variable": "col2",
                "trans": "identity",
            },
        }
    }
    assert result == expected


def test_create_graph_attrs_config_with_none_data_type():
    """Test creating graph attributes configuration with data_type=None returns inner dict."""
    result = data_handling._create_graph_attrs_config(
        column_mapping={"col1": "col1", "col2": "renamed_col2"},
        data_type=None,
        table_name="test_table",
        transformation="identity",
    )

    expected = {
        "col1": {
            "table": "test_table",
            "variable": "col1",
            "trans": "identity",
        },
        "renamed_col2": {
            "table": "test_table",
            "variable": "col2",
            "trans": "identity",
        },
    }
    assert result == expected


def test_add_results_table_to_graph(sbml_dfs_glucose_metabolism):
    """Test adding results table to graph."""
    # Create a test graph using create_napistu_graph
    graph = net_create.create_napistu_graph(
        sbml_dfs_glucose_metabolism, directed=True, wiring_approach="regulatory"
    )

    # Add some test data to sbml_dfs
    test_data = pd.DataFrame(
        {"test_attr": [1.0, 2.0, 3.0]},
        index=pd.Index(
            list(sbml_dfs_glucose_metabolism.species.index[:3]), name="s_id"
        ),
    )
    sbml_dfs_glucose_metabolism.add_species_data("test_table", test_data)

    # Test basic case - single attribute
    result = data_handling.add_results_table_to_graph(
        napistu_graph=graph,
        sbml_dfs=sbml_dfs_glucose_metabolism,
        attribute_names=["test_attr"],
        table_name="test_table",
        inplace=False,
    )
    assert "test_attr" in result.vs.attributes()

    # Test with transformation
    def square(x):
        return x**2

    result = data_handling.add_results_table_to_graph(
        napistu_graph=graph,
        sbml_dfs=sbml_dfs_glucose_metabolism,
        attribute_names=["test_attr"],
        table_name="test_table",
        transformation="square",
        custom_transformations={"square": square},
        inplace=False,
    )
    assert "test_attr" in result.vs.attributes()

    # Test inplace=True
    original_graph = graph.copy()
    assert (
        "test_attr" not in original_graph.vs.attributes()
    )  # Verify attribute doesn't exist before
    result = data_handling.add_results_table_to_graph(
        napistu_graph=original_graph,
        sbml_dfs=sbml_dfs_glucose_metabolism,
        attribute_names=["test_attr"],
        table_name="test_table",
        inplace=True,
    )
    assert result is None  # Function should return None when inplace=True
    assert (
        "test_attr" in original_graph.vs.attributes()
    )  # Verify original graph was modified

    # Test error cases
    with pytest.raises(ValueError, match="Invalid table_type"):
        data_handling.add_results_table_to_graph(
            napistu_graph=graph,
            sbml_dfs=sbml_dfs_glucose_metabolism,
            table_type="invalid",
        )

    with pytest.raises(NotImplementedError, match="Reactions are not yet supported"):
        data_handling.add_results_table_to_graph(
            napistu_graph=graph,
            sbml_dfs=sbml_dfs_glucose_metabolism,
            table_type="reactions",
        )



from typing import Optional

import click
from pydantic import BaseModel, Field, computed_field, field_validator

from napistu.mcp.constants import MCP_DEFAULTS, MCP_PRODUCTION_URL, PRESET_NAMES


class MCPServerConfig(BaseModel):
    """Server-side MCP configuration with validation."""

    host: str = Field(description="Host to bind server to")
    port: int = Field(ge=1, le=65535, description="Port to bind server to")
    server_name: str = Field(description="Server name")

    @field_validator("host")
    @classmethod
    def validate_host(cls, v):
        """Basic host validation."""
        if not v or v.isspace():
            raise ValueError("Host cannot be empty")
        return v.strip()

    @computed_field
    @property
    def bind_address(self) -> str:
        """Full address for server binding."""
        return f"{self.host}:{self.port}"


class MCPClientConfig(BaseModel):
    """Client-side MCP configuration with validation."""

    host: str = Field(description="Target server host")
    port: int = Field(ge=1, le=65535, description="Target server port")
    use_https: bool = Field(description="Use HTTPS instead of HTTP")

    @computed_field
    @property
    def base_url(self) -> str:
        """Base URL for HTTP requests."""
        protocol = "https" if self.use_https else "http"
        return f"{protocol}://{self.host}:{self.port}"

    @computed_field
    @property
    def mcp_url(self) -> str:
        """Full MCP endpoint URL."""
        return f"{self.base_url}{MCP_DEFAULTS.MCP_PATH}"


# Preset configurations - no overrides needed
def local_server_config() -> MCPServerConfig:
    """Local development server configuration."""
    return MCPServerConfig(
        host=MCP_DEFAULTS.LOCAL_HOST,
        port=MCP_DEFAULTS.LOCAL_PORT,
        server_name=MCP_DEFAULTS.LOCAL_SERVER_NAME,
    )


def production_server_config() -> MCPServerConfig:
    """Production server configuration."""
    return MCPServerConfig(
        host=MCP_DEFAULTS.PRODUCTION_HOST,
        port=MCP_DEFAULTS.PRODUCTION_PORT,
        server_name=MCP_DEFAULTS.PRODUCTION_SERVER_NAME,
    )


def local_client_config() -> MCPClientConfig:
    """Local development client configuration."""
    return MCPClientConfig(
        host=MCP_DEFAULTS.LOCAL_HOST, port=MCP_DEFAULTS.LOCAL_PORT, use_https=False
    )


def production_client_config() -> MCPClientConfig:
    """Production client configuration."""
    # Parse production URL to extract host
    production_host = MCP_PRODUCTION_URL.replace("https://", "").split("/")[0]
    return MCPClientConfig(
        host=production_host, port=MCP_DEFAULTS.HTTPS_PORT, use_https=True
    )


# Utility functions for CLI validation
def validate_server_config_flags(
    local: bool,
    production: bool,
    host: Optional[str],
    port: Optional[int],
    server_name: Optional[str],
) -> MCPServerConfig:
    """
    Validate server configuration flags and return appropriate config.

    Raises click.BadParameter if incompatible flags are used.
    """
    # Check for mutually exclusive presets
    if local and production:
        raise click.BadParameter("Cannot use both --local and --production flags")

    # Check for preset + manual config conflicts
    preset_used = local or production
    manual_config_used = any(
        [host is not None, port is not None, server_name is not None]
    )

    if preset_used and manual_config_used:
        preset_name = PRESET_NAMES.LOCAL if local else PRESET_NAMES.PRODUCTION
        raise click.BadParameter(
            f"Cannot use --{preset_name} with manual host/port/server-name configuration. "
            f"Use either preset flags OR manual configuration, not both."
        )

    # Return appropriate config
    if local:
        return local_server_config()
    elif production:
        return production_server_config()
    else:
        # Manual configuration - all required
        if not all([host, port, server_name]):
            raise click.BadParameter(
                "Manual configuration requires --host, --port, and --server-name"
            )
        return MCPServerConfig(host=host, port=port, server_name=server_name)


def validate_client_config_flags(
    local: bool,
    production: bool,
    host: Optional[str],
    port: Optional[int],
    use_https: Optional[bool],
) -> MCPClientConfig:
    """
    Validate client configuration flags and return appropriate config.

    Raises click.BadParameter if incompatible flags are used.
    """
    # Check for mutually exclusive presets
    if local and production:
        raise click.BadParameter("Cannot use both --local and --production flags")

    # Check for preset + manual config conflicts
    preset_used = local or production
    manual_config_used = any(
        [host is not None, port is not None, use_https is not None]
    )

    if preset_used and manual_config_used:
        preset_name = PRESET_NAMES.LOCAL if local else PRESET_NAMES.PRODUCTION
        raise click.BadParameter(
            f"Cannot use --{preset_name} with manual host/port/https configuration. "
            f"Use either preset flags OR manual configuration, not both."
        )

    # Return appropriate config
    if local:
        return local_client_config()
    elif production:
        return production_client_config()
    else:
        # Manual configuration - all required
        if not all([host is not None, port is not None, use_https is not None]):
            raise click.BadParameter(
                "Manual configuration requires --host, --port, and --https"
            )
        return MCPClientConfig(host=host, port=port, use_https=use_https)


# Click option decorators for reuse
def server_config_options(f):
    """Decorator to add server configuration options to click commands."""
    f = click.option(
        "--production", is_flag=True, help="Use production server configuration"
    )(f)
    f = click.option(
        "--local", is_flag=True, help="Use local development server configuration"
    )(f)
    f = click.option(
        "--host",
        type=str,
        help="Host to bind server to (requires --port and --server-name)",
    )(f)
    f = click.option(
        "--port",
        type=int,
        help="Port to bind server to (requires --host and --server-name)",
    )(f)
    f = click.option(
        "--server-name", type=str, help="Server name (requires --host and --port)"
    )(f)
    return f


def client_config_options(f):
    """Decorator to add client configuration options to click commands."""
    f = click.option("--production", is_flag=True, help="Connect to production server")(
        f
    )
    f = click.option(
        "--local", is_flag=True, help="Connect to local development server"
    )(f)
    f = click.option(
        "--host",
        type=str,
        default=None,
        help="Server host (requires --port and --https)",
    )(f)
    f = click.option(
        "--port",
        type=int,
        default=None,
        help="Server port (requires --host and --https)",
    )(f)
    f = click.option(
        "--https",
        is_flag=True,
        default=None,
        flag_value=True,
        help="Use HTTPS (requires --host and --port)",
    )(f)
    return f



import logging
from types import SimpleNamespace
from typing import Dict

from napistu.constants import ONTOLOGIES
from napistu.ontologies._validation import SpeciesTypeOntologyMapping

logger = logging.getLogger(__name__)

# Valid ontologies that can be interconverted
INTERCONVERTIBLE_GENIC_ONTOLOGIES = {
    ONTOLOGIES.ENSEMBL_GENE,
    ONTOLOGIES.ENSEMBL_TRANSCRIPT,
    ONTOLOGIES.ENSEMBL_PROTEIN,
    ONTOLOGIES.NCBI_ENTREZ_GENE,
    ONTOLOGIES.UNIPROT,
    ONTOLOGIES.GENE_NAME,
    ONTOLOGIES.SYMBOL,
}

GENODEXITO_DEFS = SimpleNamespace(
    BIOCONDUCTOR="bioconductor",
    PYTHON="python",
)
GENODEXITO_MAPPERS = {GENODEXITO_DEFS.BIOCONDUCTOR, GENODEXITO_DEFS.PYTHON}

# Mapping from our ontology names to MyGene field names
MYGENE_DEFS = SimpleNamespace(
    ENSEMBL_GENE="ensembl.gene",
    ENSEMBL_TRANSCRIPT="ensembl.transcript",
    ENSEMBL_PROTEIN="ensembl.protein",
    UNIPROT="uniprot.Swiss-Prot",
    SYMBOL="symbol",
    GENE_NAME="name",
    NCBI_ENTREZ_GENE="entrezgene",
)

NAPISTU_TO_MYGENE_FIELDS = {
    ONTOLOGIES.ENSEMBL_GENE: MYGENE_DEFS.ENSEMBL_GENE,
    ONTOLOGIES.ENSEMBL_TRANSCRIPT: MYGENE_DEFS.ENSEMBL_TRANSCRIPT,
    ONTOLOGIES.ENSEMBL_PROTEIN: MYGENE_DEFS.ENSEMBL_PROTEIN,
    ONTOLOGIES.UNIPROT: MYGENE_DEFS.UNIPROT,
    ONTOLOGIES.SYMBOL: MYGENE_DEFS.SYMBOL,
    ONTOLOGIES.GENE_NAME: MYGENE_DEFS.GENE_NAME,
    ONTOLOGIES.NCBI_ENTREZ_GENE: MYGENE_DEFS.NCBI_ENTREZ_GENE,
}

NAPISTU_FROM_MYGENE_FIELDS = {
    MYGENE_DEFS.ENSEMBL_GENE: ONTOLOGIES.ENSEMBL_GENE,
    MYGENE_DEFS.ENSEMBL_TRANSCRIPT: ONTOLOGIES.ENSEMBL_TRANSCRIPT,
    MYGENE_DEFS.ENSEMBL_PROTEIN: ONTOLOGIES.ENSEMBL_PROTEIN,
    MYGENE_DEFS.UNIPROT: ONTOLOGIES.UNIPROT,
    MYGENE_DEFS.SYMBOL: ONTOLOGIES.SYMBOL,
    MYGENE_DEFS.GENE_NAME: ONTOLOGIES.GENE_NAME,
    MYGENE_DEFS.NCBI_ENTREZ_GENE: ONTOLOGIES.NCBI_ENTREZ_GENE,
}

SPECIES_TO_TAXID: Dict[str, int] = {
    # MyGene.info supported common species (9 species with common names)
    "Homo sapiens": 9606,  # human
    "Mus musculus": 10090,  # mouse
    "Rattus norvegicus": 10116,  # rat
    "Drosophila melanogaster": 7227,  # fruitfly
    "Caenorhabditis elegans": 6239,  # nematode
    "Danio rerio": 7955,  # zebrafish
    "Arabidopsis thaliana": 3702,  # thale-cress
    "Xenopus tropicalis": 8364,  # frog
    "Xenopus laevis": 8355,  # frog (alternative species)
    "Sus scrofa": 9823,  # pig
    # Additional commonly used model organisms
    "Saccharomyces cerevisiae": 4932,  # yeast
    "Schizosaccharomyces pombe": 4896,  # fission yeast
    "Gallus gallus": 9031,  # chicken
    "Bos taurus": 9913,  # cow/cattle
    "Canis familiaris": 9615,  # dog
    "Macaca mulatta": 9544,  # rhesus monkey/macaque
    "Pan troglodytes": 9598,  # chimpanzee
    "Escherichia coli": 511145,  # E. coli (K-12 MG1655)
    # Additional species that might be encountered
    "Anopheles gambiae": 7165,  # malaria mosquito
    "Oryza sativa": 4530,  # rice
    "Neurospora crassa": 5141,  # bread mold
    "Kluyveromyces lactis": 28985,  # yeast species
    "Magnaporthe oryzae": 318829,  # rice blast fungus
    "Eremothecium gossypii": 33169,  # cotton fungus
}

MYGENE_QUERY_DEFS = SimpleNamespace(
    BIOLOGICAL_REGION="type_of_gene:biological-region",
    NCRNA="type_of_gene:ncrna",
    PROTEIN_CODING="type_of_gene:protein-coding",
    PSEUDO="type_of_gene:pseudo",
    SNORNA="type_of_gene:snorna",
    UNKNOWN="type_of_gene:unknown",
    OTHER="type_of_gene:other",
    RRNA="type_of_gene:rrna",
    TRNA="type_of_gene:trna",
    SNRNA="type_of_gene:snrna",
)

MYGENE_QUERY_DEFS_LIST = list(MYGENE_QUERY_DEFS.__dict__.values())

MYGENE_DEFAULT_QUERIES = [MYGENE_QUERY_DEFS.PROTEIN_CODING, MYGENE_QUERY_DEFS.NCRNA]

# bioc ontologies used for linking systematic identifiers
# (entrez is not part of this list because it forms the gene index)
PROTEIN_ONTOLOGIES = [ONTOLOGIES.UNIPROT, ONTOLOGIES.ENSEMBL_PROTEIN]
GENE_ONTOLOGIES = [
    ONTOLOGIES.NCBI_ENTREZ_GENE,
    ONTOLOGIES.ENSEMBL_GENE,
    ONTOLOGIES.ENSEMBL_TRANSCRIPT,
]
NAME_ONTOLOGIES = {
    ONTOLOGIES.GENE_NAME: 0,
    ONTOLOGIES.SYMBOL: 1,
    ONTOLOGIES.UNIPROT: 2,
    ONTOLOGIES.ENSEMBL_PROTEIN: 3,
}

# PubChem constants
PUBCHEM_ID_ENTRYPOINT = "https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/cid/{cids_str}/property/IUPACName,Title,IsomericSMILES,SMILES/JSON"

PUBCHEM_DEFS = SimpleNamespace(NAME="name", SMILES="smiles", PUBCHEM_ID="pubchem_id")

PUBCHEM_PROPERTIES = SimpleNamespace(
    CID="CID",
    IUPAC_NAME="IUPACName",
    TITLE="Title",
    ISOMERIC_SMILES="IsomericSMILES",
    SMILES="SMILES",
    PROPERTY_TABLE="PropertyTable",
    PROPERTIES="Properties",
)

# miRBase constants

MIRBASE_TABLES = SimpleNamespace(
    MATURE_DATABASE_LINKS="mature_database_links",
    MATURE_DATABASE_URL="mature_database_url",
    URL="url",
    HEADER="header",
    DATABASE_ENTRY="database_entry",
    DATABASE="database",
    URL_TEMPLATE="url_template",  # unused
    RNA_ID="rna_id",  # distinct molecules
    PRIMARY_ID="primary_id",
    SECONDARY_ID="secondary_id",
)

MIRBASE_TABLE_SPECS = {
    MIRBASE_TABLES.MATURE_DATABASE_LINKS: {
        MIRBASE_TABLES.URL: "https://mirbase.org/download/CURRENT/database_files/mature_database_links.txt",
        MIRBASE_TABLES.HEADER: [
            MIRBASE_TABLES.RNA_ID,
            MIRBASE_TABLES.DATABASE_ENTRY,
            MIRBASE_TABLES.PRIMARY_ID,
            MIRBASE_TABLES.SECONDARY_ID,
        ],
    },
    MIRBASE_TABLES.MATURE_DATABASE_URL: {
        MIRBASE_TABLES.URL: "https://mirbase.org/download/CURRENT/database_files/mature_database_url.txt",
        MIRBASE_TABLES.HEADER: [
            MIRBASE_TABLES.DATABASE_ENTRY,
            MIRBASE_TABLES.DATABASE,
            MIRBASE_TABLES.URL_TEMPLATE,
            "unknown",
        ],
    },
}

# Add your species type mappings

SPECIES_TYPES = SimpleNamespace(
    COMPLEX="complex",
    DRUG="drug",
    METABOLITE="metabolite",
    PROTEIN="protein",
    REGULATORY_RNA="regulatory_rna",
    OTHER="other",
    UNKNOWN="unknown",
)

SPECIES_TYPE_PLURAL = {
    SPECIES_TYPES.COMPLEX: "complexes",
    SPECIES_TYPES.DRUG: "drugs",
    SPECIES_TYPES.METABOLITE: "metabolites",
    SPECIES_TYPES.PROTEIN: "proteins",
    SPECIES_TYPES.REGULATORY_RNA: "regulatory RNAs",
    SPECIES_TYPES.OTHER: "other",
    SPECIES_TYPES.UNKNOWN: "unknowns",
}

SPECIES_TYPE_ONTOLOGIES = {
    SPECIES_TYPES.COMPLEX: [ONTOLOGIES.CORUM],
    SPECIES_TYPES.DRUG: [ONTOLOGIES.DRUGBANK, ONTOLOGIES.KEGG_DRUG],
    SPECIES_TYPES.METABOLITE: [
        ONTOLOGIES.BIGG_METABOLITE,
        ONTOLOGIES.CHEBI,
        ONTOLOGIES.KEGG,
        ONTOLOGIES.PUBCHEM,
        ONTOLOGIES.SMILES,
    ],
    SPECIES_TYPES.PROTEIN: [
        ONTOLOGIES.ENSEMBL_GENE,
        ONTOLOGIES.ENSEMBL_TRANSCRIPT,
        ONTOLOGIES.ENSEMBL_PROTEIN,
        ONTOLOGIES.NCBI_ENTREZ_GENE,
        ONTOLOGIES.UNIPROT,
        ONTOLOGIES.SYMBOL,
        ONTOLOGIES.GENE_NAME,
    ],
    SPECIES_TYPES.REGULATORY_RNA: [ONTOLOGIES.MIRBASE, ONTOLOGIES.RNACENTRAL],
}

# if the ontology's associated with these categories are seen then other categories are ignored
PRIORITIZED_SPECIES_TYPES = {SPECIES_TYPES.DRUG, SPECIES_TYPES.COMPLEX}

# Validate the mapping and create the flattened lookup at module load time
validated_mapping = SpeciesTypeOntologyMapping(mappings=SPECIES_TYPE_ONTOLOGIES)
ONTOLOGY_TO_SPECIES_TYPE = validated_mapping.create_ontology_to_species_type_mapping()



"""
web_routes.py - Route handlers for Napistu chat web interface with CORS support
"""

import logging
from typing import Optional, Tuple

import httpx
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import JSONResponse, RedirectResponse
from starlette.routing import Route

from napistu.mcp.chat_web import (
    get_chat_config,
    get_claude_client,
    get_cost_tracker,
    get_rate_limiter,
)
from napistu.mcp.constants import DEFAULT_ALLOWED_ORIGINS, MCP_PRODUCTION_URL

logger = logging.getLogger(__name__)


class ChatMessage(BaseModel):
    content: str


def create_chat_app() -> Starlette:
    """
    Create a Starlette app for chat routes with CORS middleware.

    This app is completely separate from the MCP server and only handles
    the /api/* chat endpoints. CORS is only applied to these routes.

    Returns
    -------
    Starlette
        Starlette app with chat routes and CORS middleware
    """

    chat_app = Starlette(
        routes=[
            Route("/api/chat", endpoint=handle_chat, methods=["POST", "OPTIONS"]),
            Route("/api/stats", endpoint=handle_stats, methods=["GET", "OPTIONS"]),
            Route("/api/health", endpoint=handle_health, methods=["GET", "OPTIONS"]),
            Route("/api/test-mcp", endpoint=handle_mcp_test, methods=["GET"]),
        ]
    )

    # Add CORS middleware ONLY to this app
    chat_app.add_middleware(
        CORSMiddleware,
        allow_origins=DEFAULT_ALLOWED_ORIGINS,
        allow_credentials=True,
        allow_methods=["GET", "POST", "OPTIONS"],
        allow_headers=["Content-Type"],
    )

    logger.info("Created chat API with CORS middleware at /api/*")

    return chat_app


async def handle_chat(request: Request) -> JSONResponse:
    """Handle chat requests with rate limiting and cost tracking"""
    # debugging
    try:
        # Get client IP
        ip = request.client.host
        logger.info(f"Chat request from IP: {ip}")

        # Parse and validate request body
        message, error = await _parse_chat_message(request)
        if error:
            return error

        # Initialize all components (lazy init with fail-fast)
        success, error = _initialize_chat_components()
        if not success:
            return error

        # Get initialized components
        config = get_chat_config()
        rate_limiter = get_rate_limiter()
        cost_tracker = get_cost_tracker()
        client = get_claude_client()

        # Validate request against all constraints
        error = _validate_chat_request(message, ip, config, rate_limiter, cost_tracker)
        if error:
            return error

        # Call Claude with MCP tools
        logger.info(f"Calling Claude API for message: {message.content[:50]}...")
        try:
            result = await client.chat(message.content)
            logger.info(f"Claude API response received: {result['usage']}")
        except Exception as e:
            logger.error(f"Claude API call failed: {type(e).__name__}: {e}")
            import traceback

            logger.error(traceback.format_exc())
            return JSONResponse(
                content={
                    "detail": "Failed to get response from Claude API",
                    "error": str(e),
                },
                status_code=500,
            )

        # Record usage
        rate_limiter.record_request(ip)
        cost_tracker.record_cost(result["usage"])
        logger.info(f"Request completed successfully for IP {ip}")

        return JSONResponse(content=result)

    except Exception as e:
        logger.error(f"Unexpected error in handle_chat: {type(e).__name__}: {e}")
        import traceback

        logger.error(traceback.format_exc())
        return JSONResponse(
            content={
                "detail": f"Internal error: {str(e)}",
                "error_type": type(e).__name__,
            },
            status_code=500,
        )


async def handle_health(request: Request) -> JSONResponse:
    """Health check for chat API"""
    try:
        client = get_claude_client()
        api_configured = client.client is not None
        cost_tracker = get_cost_tracker()

        return JSONResponse(
            content={
                "status": "healthy",
                "chat_api": "configured" if api_configured else "not_configured",
                "budget_ok": cost_tracker.check_budget(),
            },
        )
    except Exception as e:
        return JSONResponse(
            content={"status": "unhealthy", "error": str(e)},
            status_code=500,
        )


async def handle_mcp_test(request: Request) -> JSONResponse:
    """Test if MCP server is reachable from this container"""

    mcp_url = MCP_PRODUCTION_URL + "/mcp/"

    try:
        async with httpx.AsyncClient(timeout=10.0) as client:
            response = await client.post(
                mcp_url,
                headers={
                    "Content-Type": "application/json",
                    "Accept": "application/json, text/event-stream",
                },
                json={
                    "jsonrpc": "2.0",
                    "id": 1,
                    "method": "initialize",
                    "params": {
                        "protocolVersion": "2024-11-05",
                        "capabilities": {},
                        "clientInfo": {"name": "test", "version": "1.0"},
                    },
                },
            )
            return JSONResponse(
                content={
                    "status": "reachable",
                    "code": response.status_code,
                    "can_reach_external_mcp": True,
                    "url_tested": mcp_url,
                }
            )
    except Exception as e:
        return JSONResponse(
            content={
                "status": "unreachable",
                "error": str(e),
                "can_reach_external_mcp": False,
                "url_tested": mcp_url,
            },
            status_code=500,
        )


async def handle_stats(request: Request) -> JSONResponse:
    """Get current usage stats"""
    try:
        # Get client IP for rate limit calculations
        ip = request.client.host

        chat_config = get_chat_config()
        cost_tracker = get_cost_tracker()
        rate_limiter = get_rate_limiter()
        cost_stats = cost_tracker.get_stats()

        # Get remaining requests for this IP
        remaining = rate_limiter.get_remaining_requests(ip)

        stats = {
            "budget": {
                "daily_limit": chat_config.daily_budget,
                "spent_today": cost_stats["cost_today"],
                "remaining": cost_stats["budget_remaining"],
            },
            "rate_limits": remaining,  # Now just uses the method directly
        }

        logger.info(
            f"Stats for IP {ip}: {remaining['per_hour']} remaining/hour, "
            f"{remaining['per_day']} remaining/day"
        )

        return JSONResponse(content=stats)
    except Exception as e:
        return JSONResponse(
            content={"detail": f"Error getting stats: {str(e)}"},
            status_code=500,
        )


async def redirect_to_mcp(request: Request) -> RedirectResponse:
    """Redirect /mcp to /mcp/ for trailing slash compatibility"""
    return RedirectResponse(url="/mcp/", status_code=307)


# private utils


def _initialize_chat_components() -> Tuple[bool, Optional[JSONResponse]]:
    """
    Initialize all chat components with lazy loading and error handling.

    Returns
    -------
    Tuple[bool, Optional[JSONResponse]]
        (success, error_response)
    """
    try:
        _ = get_chat_config()
        _ = get_rate_limiter()
        _ = get_cost_tracker()
        _ = get_claude_client()
        return True, None
    except Exception as e:
        logger.error(f"Failed to initialize chat components: {e}")
        error_response = JSONResponse(
            content={
                "detail": "Chat service configuration error. Please check server logs.",
                "error_type": type(e).__name__,
            },
            status_code=503,
        )
        return False, error_response


async def _parse_chat_message(
    request: Request,
) -> Tuple[Optional[ChatMessage], Optional[JSONResponse]]:
    """
    Parse and validate chat message from request body.

    Returns
    -------
    Tuple[Optional[ChatMessage], Optional[JSONResponse]]
        (message, error_response) where one is always None
    """
    try:
        body = await request.json()
        message = ChatMessage(**body)
        return message, None
    except Exception as e:
        logger.error(f"Failed to parse request body: {e}")
        error_response = JSONResponse(
            content={"detail": f"Invalid request body: {str(e)}"},
            status_code=400,
        )
        return None, error_response


def _validate_chat_request(
    message: ChatMessage,
    ip: str,
    config,
    rate_limiter,
    cost_tracker,
) -> Optional[JSONResponse]:
    """
    Validate chat request against all constraints.

    Returns
    -------
    Optional[JSONResponse]
        Error response if validation fails, None if all checks pass
    """
    # Validate message length
    if not message.content or len(message.content) > config.max_message_length:
        logger.warning(
            f"Message length validation failed: {len(message.content)} chars"
        )
        return JSONResponse(
            content={
                "detail": f"Message must be between 1 and {config.max_message_length} characters."
            },
            status_code=400,
        )

    # Check rate limits
    is_allowed, error_msg = rate_limiter.check_limit(ip)
    if not is_allowed:
        logger.warning(f"Rate limit exceeded for IP {ip}: {error_msg}")
        return JSONResponse(
            content={"detail": error_msg},
            status_code=429,
        )

    # Check daily budget
    if not cost_tracker.check_budget():
        logger.warning("Daily budget exceeded")
        return JSONResponse(
            content={
                "detail": "Daily budget exceeded. Service will be available again tomorrow."
            },
            status_code=503,
        )

    return None
